(in-package "DEDUC")

(def-pddl-form-handler context definer (def-form)
   (multiple-value-bind (status name dom fields flg-junk)
                        (massage-domain-entity-def-form def-form)
      (case status
         (:okay
	  (let ((obs (collect-field ':objects fields))
		(inits (append-field ':init fields))
		(bad-fields (collect-bad-fields '(:domain :objects :init)
						fields)))
	     (context-parse name dom obs inits
	          `(,@bad-fields ,@flg-junk))))
	 (t
	  (note-domain-entity-bogosity status name dom fields flg-junk
                                       def-form)))))


(defun context-parse (name dom obs inits flg-junk)
      (domain-make-current dom)
      (multiple-value-bind (sit flg-consts flg-inits)
			   (build-context name (list dom) obs inits)
	 `(define (context ,(cond ((situation-p sit) name)
				    (t (flag-wrong-type
						sit name 'situation))))
	     (:domain ,(domain-name dom))
	     ,@flg-consts
	     (:init ,@flg-inits)
	     ,@flg-junk)))

(declaim (special flag-count*))

(defun build-context (name doms obs inits)
   (let ((idom (domain-indexify (car doms))))
      (let ((sit (domain-place-situation idom name))
	    (initial-flag-count flag-count*))
	 (let ((subdom (domain-indexify (initial-situation-domain sit))))
	   (setf (domain-parents subdom) (mapcar #'domain-indexify doms))
	   (domain-reset subdom)
	   (reset-ancestors subdom)
	   (let ((flg-consts (constant-lists-parse ':objects obs subdom)))
	       (multiple-value-bind (props prop-flg-junk)
				    (list-smooth inits #'consp)
		  (let ((lits (mapcar #'(lambda (a)
					    (formula-parse a '() t 'notrans subdom))
				      props)))
		     (setf (initial-situation-delta sit) lits)
		     (cond ((= flag-count* initial-flag-count)
			    (situation-initialize sit)
			    (context-forward-chain sit)))
		     (values sit flg-consts (append lits prop-flg-junk)))))))))


(defun context-forward-chain (con)
   (let ((cdom (initial-situation-domain con)))
      (indexed-domain-make-current cdom)
      (dolist (occ (exp_index-contents (get-situation-index con) nil))
	 (trigger-forward-rules (occasion-prop occ) dummy-id* '() con cdom))
      (dolist (occ (exp_index-contents (initial-situation-timeless-index con) nil))
	 (trigger-forward-rules (occasion-prop occ) dummy-id* '() con cdom))))


(defun new-deduction-context (old add delete dom)
   (setq dom (domain-indexify dom))
   (indexed-domain-make-current dom)
   (let (predecessor new-index)
      (cond (old
	     (setq predecessor (find-init-situation old))
	     (setq new-index (exp-index-copy (get-situation-index old))))
	    (t
	     (setq predecessor (domain-place-timeless-sit dom))
	     (setq new-index (exp-index-init #'occasion-prop))))
      (let ((newsit
	       (make-subsequent-situation
		  :index new-index
		  :contents '*variable
		  :init predecessor
		  :path (cons 'push-context (cond (old (situation-path old))
						  (t '()))))))
	 (dolist (d delete)
	     (assertion-erase d dummy-id* '() newsit dom))
	 (dolist (a add)
	    (assertion-add a dummy-id* '() newsit dom)))))

;;; If we start an anonymous context from scratch, we have to create an initial
;;; situation to hold a timeless index (and any conclusions drawn purely from
;;; timeless assertions).
(defun domain-place-timeless-sit (dom)
   (let ((placeholder-sit 
	    (domain-place-situation dom (intern (format nil "~a-TIMELESS"
							(domain-name dom))))))
      (situation-initialize placeholder-sit)
      (context-forward-chain placeholder-sit)
      placeholder-sit))

(defun find-named-context (name dom)
   (let ((bdg (find-domain-bdg name dom)))
     (cond (bdg
	    (let ((c (pddl-bdg-val bdg)))
	       (cond ((initial-situation-p c)
		      c)
		     (t
		      (error "~s is not a situation of ~s [value: ~s]"
			     name dom c)))))
	   (t
	    (error "~s unbound in ~s"
		   name dom)))))

(defun all-assertions (con)
   (nconc (exp_index-contents (get-situation-index con) t)
	  (exp_index-contents (initial-situation-timeless-index
			          (find-init-situation con))
			      t)))

(defun all-axioms (dom)
   (exp_index-contents
       (indexed-domain-axioms dom)
       t))
  




